@@ -454,87 +454,121 @@ void add(GF2X& x, const GF2X& a, const GF2X& b)
 
 
 
-
-// This mul1 routine (for 32 x 32 bit multiplies) 
-// was arrived at after a good deal of experimentation.
-// Several people have advocated that the "best" way
-// is to reduce it via karastuba to 9 8x8 multiplies, implementing
-// the latter via table look-up (a huge table).  Although such a mul1
-// would indeed have fewer machine instructions, my
-// experiments on a PowerPC and on a Pentium indicate
-// that the memory accesses slow it down.  On these machines
-// the following mul1 is faster by a factor of about 1.5.
+/*
+ * The bodies of mul1, Mul1, AddMul1, and mul_half
+ * are generated by the MakeDesc program, and the
+ * macros NTL_BB_MUL_CODE... are defined in mach_desc.h.
+ * Thanks to Paul Zimmermann for providing improvements
+ * to this approach.
+ */
 
 
-// inlining mul1 seems to help with g++; morever,
-// the IBM xlC compiler inlines it anyway.
+#if (defined(NTL_GF2X_ALTCODE1))
 
-static inline
+#define NTL_EFF_BB_MUL_CODE0 NTL_ALT1_BB_MUL_CODE0
+#define NTL_EFF_BB_MUL_CODE1 NTL_ALT1_BB_MUL_CODE1
+#define NTL_EFF_BB_MUL_CODE2 NTL_ALT1_BB_MUL_CODE2
+#define NTL_EFF_SHORT_BB_MUL_CODE1 NTL_ALT1_SHORT_BB_MUL_CODE1
+#define NTL_EFF_HALF_BB_MUL_CODE0 NTL_ALT1_HALF_BB_MUL_CODE0
+
+#elif (defined(NTL_GF2X_ALTCODE))
+
+#define NTL_EFF_BB_MUL_CODE0 NTL_ALT_BB_MUL_CODE0
+#define NTL_EFF_BB_MUL_CODE1 NTL_ALT_BB_MUL_CODE1
+#define NTL_EFF_BB_MUL_CODE2 NTL_ALT_BB_MUL_CODE2
+#define NTL_EFF_SHORT_BB_MUL_CODE1 NTL_ALT_SHORT_BB_MUL_CODE1
+#define NTL_EFF_HALF_BB_MUL_CODE0 NTL_ALT_HALF_BB_MUL_CODE0
+
+#else
+
+#define NTL_EFF_BB_MUL_CODE0 NTL_BB_MUL_CODE0
+#define NTL_EFF_BB_MUL_CODE1 NTL_BB_MUL_CODE1
+#define NTL_EFF_BB_MUL_CODE2 NTL_BB_MUL_CODE2
+#define NTL_EFF_SHORT_BB_MUL_CODE1 NTL_SHORT_BB_MUL_CODE1
+#define NTL_EFF_HALF_BB_MUL_CODE0 NTL_HALF_BB_MUL_CODE0
+
+#endif
+
+
+
+static 
 void mul1(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)
 {
-   _ntl_ulong hi, lo;
-   _ntl_ulong A[4];
 
-   A[0] = 0;
-   A[1] = a & ((1UL << (NTL_BITS_PER_LONG-1))-1UL);
-   A[2] = a << 1;
-   A[3] = A[1] ^ A[2];
+NTL_EFF_BB_MUL_CODE0
 
-   lo = A[b>>(NTL_BITS_PER_LONG-2)];
-   hi = lo >> (NTL_BITS_PER_LONG-2); 
-   lo = (lo << 2) ^ A[(b >> (NTL_BITS_PER_LONG-4)) & 3];
 
-   // The following code is included from mach_desc.h
-   // and handles *any* word size
-
-   NTL_BB_MUL_CODE
-
-   hi = (hi << 2)|(lo >> (NTL_BITS_PER_LONG-2)); 
-   lo = (lo << 2) ^ A[b & 3];
-
-   if (a >> (NTL_BITS_PER_LONG-1)) {
-      hi = hi ^ (b >> 1);
-      lo = lo ^ (b << (NTL_BITS_PER_LONG-1));
-   }
-
-   c[0] = lo;
-   c[1] = hi;
 }
 
+
+#ifdef NTL_GF2X_NOINLINE
+
+#define mul1_IL mul1
+
+#else
+
 static inline
+void mul1_inline(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)
+{
+
+
+NTL_EFF_BB_MUL_CODE0
+
+
+}
+
+#define mul1_IL mul1_inline 
+#endif
+
+
+static 
+void Mul1(_ntl_ulong *cp, const _ntl_ulong *bp, long sb, _ntl_ulong a)
+{
+ 
+
+NTL_EFF_BB_MUL_CODE1
+
+
+}
+
+static 
+void AddMul1(_ntl_ulong *cp, const _ntl_ulong* bp, long sb, _ntl_ulong a)
+{
+
+
+NTL_EFF_BB_MUL_CODE2
+
+
+}
+
+
+static 
+void Mul1_short(_ntl_ulong *cp, const _ntl_ulong *bp, long sb, _ntl_ulong a)
+{
+ 
+
+NTL_EFF_SHORT_BB_MUL_CODE1
+
+
+}
+
+
+
+
+
+static 
 void mul_half(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)
 {
-   _ntl_ulong hi, lo;
-   _ntl_ulong A[4];
-
-   A[0] = 0;
-   A[1] = a & ((1UL << (NTL_BITS_PER_LONG-1))-1UL);
-   A[2] = a << 1;
-   A[3] = A[1] ^ A[2];
-
-   lo = A[b>>(NTL_BITS_PER_LONG/2-2)];
-   hi = lo >> (NTL_BITS_PER_LONG-2); 
-   lo = (lo << 2) ^ A[(b >> (NTL_BITS_PER_LONG/2-4)) & 3];
 
 
-   NTL_BB_HALF_MUL_CODE
+NTL_EFF_HALF_BB_MUL_CODE0
 
 
-   hi = (hi << 2)|(lo >> (NTL_BITS_PER_LONG-2)); 
-   lo = (lo << 2) ^ A[b & 3];
-
-   if (a >> (NTL_BITS_PER_LONG-1)) {
-      hi = hi ^ (b >> 1);
-      lo = lo ^ (b << (NTL_BITS_PER_LONG-1));
-   }
-
-   c[0] = lo;
-   c[1] = hi;
 }
 
 
 // mul2...mul8 hard-code 2x2...8x8 word multiplies.
-// I adapted these routines from LiDIA.
+// I adapted these routines from LiDIA (except mul3, see below).
 // Inlining these seems to hurt, not help.
 
 static
@@ -546,9 +580,9 @@ void mul2(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
    hs0 = a[0] ^ a[1];
    hs1 = b[0] ^ b[1];
 
-   mul1(c, a[0], b[0]);
-   mul1(c+2, a[1], b[1]);
-   mul1(hl2, hs0, hs1);
+   mul1_IL(c, a[0], b[0]);
+   mul1_IL(c+2, a[1], b[1]);
+   mul1_IL(hl2, hs0, hs1);
 
 
    hl2[0] = hl2[0] ^ c[0] ^ c[2];
@@ -558,30 +592,33 @@ void mul2(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
    c[2] ^= hl2[1];
 }
 
+
+/*
+ * This version of mul3 I got from Weimerskirch, Stebila, 
+ * and Shantz, "Generic GF(2^m) arithmetic in software
+ * an its application to ECC" (ACISP 2003).
+ */
+
 static
 void mul3 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
 {
-   _ntl_ulong hs0[2], hs1[2];
-   _ntl_ulong hl2[4];
+   _ntl_ulong d0[2], d1[2], d2[2], d01[2], d02[2], d12[2];
 
-   hs0[0] = a[0] ^ a[2]; 
-   hs0[1] = a[1];
-   hs1[0] = b[0] ^ b[2]; 
-   hs1[1] = b[1];
+   mul1_IL(d0, a[0], b[0]);
+   mul1_IL(d1, a[1], b[1]);
+   mul1_IL(d2, a[2], b[2]);
+   mul1_IL(d01, a[0]^a[1], b[0]^b[1]);
+   mul1_IL(d02, a[0]^a[2], b[0]^b[2]);
+   mul1_IL(d12, a[1]^a[2], b[1]^b[2]);
 
-   mul2(c, a, b); 
-   mul1(c+4, a[2], b[2]);
-   mul2(hl2, hs0, hs1); 
-
-   hl2[0] = hl2[0] ^ c[0] ^ c[4]; 
-   hl2[1] = hl2[1] ^ c[1] ^ c[5];
-   hl2[2] = hl2[2] ^ c[2];
-   hl2[3] = hl2[3] ^ c[3];
    
-   c[2] ^= hl2[0];
-   c[3] ^= hl2[1];
-   c[4] ^= hl2[2];
-   c[5] ^= hl2[3];
+   c[0] = d0[0];
+   c[1] = d0[1] ^ d01[0] ^ d1[0] ^ d0[0];
+   c[2] = d01[1] ^ d1[1] ^ d0[1] ^ d02[0] ^ d2[0] ^ d0[0] ^ d1[0];
+   c[3] = d02[1] ^ d2[1] ^ d0[1] ^ d1[1] ^ d12[0] ^ d1[0] ^ d2[0];
+   c[4] = d12[1] ^ d1[1] ^ d2[1] ^ d2[0];
+   c[5] = d2[1];
+
 }
 
 static
@@ -624,8 +661,8 @@ void mul5 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
    hs1[2] = b[2];
 
    mul3(c, a, b); 
-   mul2(c+6, a+3, b+3); 
    mul3(hl2, hs0, hs1);
+   mul2(c+6, a+3, b+3);
 
    hl2[0] = hl2[0] ^ c[0] ^ c[6]; 
    hl2[1] = hl2[1] ^ c[1] ^ c[7];
@@ -691,8 +728,8 @@ void mul7(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
    hs1[3] = b[3];
 
    mul4(c, a, b); 
-   mul3(c+8, a+4, b+4); 
    mul4(hl2, hs0, hs1);
+   mul3(c+8, a+4, b+4); 
 
    hl2[0] = hl2[0] ^ c[0] ^ c[8];
    hl2[1] = hl2[1] ^ c[1] ^ c[9];
@@ -774,6 +811,7 @@ void KarMul(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
    const _ntl_ulong *a0, *a1, *b0, *b1;
    _ntl_ulong *a01, *b01, *h;
 
+
    lh = len >> 1;
    ll = (len+1) >> 1;
 
@@ -815,6 +853,7 @@ void KarMul(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
 }
 
 
+
 void mul(GF2X& c, const GF2X& a, const GF2X& b)
 {
    long sa = a.xrep.length();
@@ -1105,35 +1144,19 @@ void mul(GF2X& c, const GF2X& a, const GF2X& b)
    }
 
    // another special case:  one of the two inputs
-   // has length 1 (and maybe 1/2).
+   // has length 1 (or less).
 
    if (sa == 1) {
       c.xrep.SetLength(sb + 1);
       _ntl_ulong *cp = c.xrep.elts();
       const _ntl_ulong *bp = b.xrep.elts();
 
-      long i;
-      _ntl_ulong carry;
-      _ntl_ulong v[2];
+      if (a0 >> (NTL_BITS_PER_LONG-NTL_BB_MUL1_BITS+1))
+         Mul1(cp, bp, sb, a0);
+      else
+         Mul1_short(cp, bp, sb, a0);
 
-      carry = 0;
 
-      if (a0 >> NTL_BITS_PER_LONG/2) {
-         for (i = 0; i < sb; i++) {
-            mul1(v, bp[i], a0);
-            cp[i] = carry ^ v[0];
-            carry = v[1];
-         }
-      }
-      else {
-         for (i = 0; i < sb; i++) {
-            mul_half(v, bp[i], a0);
-            cp[i] = carry ^ v[0];
-            carry = v[1];
-         }
-      }
-
-      cp[sb] = carry;
       c.normalize();
       return;
    }
@@ -1143,28 +1166,12 @@ void mul(GF2X& c, const GF2X& a, const GF2X& b)
       _ntl_ulong *cp = c.xrep.elts();
       const _ntl_ulong *ap = a.xrep.elts();
 
-      long i;
-      _ntl_ulong carry;
-      _ntl_ulong v[2];
 
-      carry = 0;
+      if (b0 >> (NTL_BITS_PER_LONG-NTL_BB_MUL1_BITS+1))
+         Mul1(cp, ap, sa, b0);
+      else
+         Mul1_short(cp, ap, sa, b0);
 
-      if (b0 >> NTL_BITS_PER_LONG/2) {
-         for (i = 0; i < sa; i++) {
-            mul1(v, ap[i], b0);
-            cp[i] = carry ^ v[0];
-            carry = v[1];
-         }
-      }
-      else {
-         for (i = 0; i < sa; i++) {
-            mul_half(v, ap[i], b0);
-            cp[i] = carry ^ v[0];
-            carry = v[1];
-         }
-      }
-
-      cp[sa] = carry;
       c.normalize();
       return;
    }
@@ -1231,12 +1238,8 @@ void mul(GF2X& c, const GF2X& a, const GF2X& b)
       if (sa == 0) break;
 
       if (sa == 1) {
-         for (i = 0; i < sb; i++) {
-            mul1(v, ap[0], bp[i]);
-            cp[i] ^= v[0];
-            cp[i+1] ^= v[1];
-         }
-
+         AddMul1(cp, bp, sb, ap[0]);
+         
          break;
       }
 
@@ -1261,6 +1264,7 @@ void mul(GF2X& c, const GF2X& a, const GF2X& b)
 
 
 
+
 void mul(GF2X& c, const GF2X& a, long b)
 {
    if (b & 1)
