@@ -8,6 +8,22 @@
 
 NTL_START_IMPL
 
+
+static inline
+void CheckFinite(double *p)
+{
+   if (!IsFinite(p)) Error("G_LLL_QP: numbers too big...use G_LLL_XD");
+}
+
+
+static inline
+void CheckFinite(quad_float *p)
+{
+   if (!IsFinite(p)) Error("G_LLL_QP: numbers too big...use G_LLL_XD");
+}
+
+
+
 static void RowTransform(vec_ZZ& A, vec_ZZ& B, const ZZ& MU1)
 // x = x - y*MU
 {
@@ -63,15 +79,15 @@ static void RowTransform(vec_ZZ& A, vec_ZZ& B, const ZZ& MU1)
 #define TR_BND (NTL_FDOUBLE_PRECISION/2.0)
 // Just to be safe!!
 
-static quad_float max_abs(quad_float *v, long n)
+static double max_abs(quad_float *v, long n)
 {
    long i;
-   quad_float res, t;
+   double res, t;
 
    res = 0;
 
    for (i = 1; i <= n; i++) {
-      t = fabs(v[i]);
+      t = fabs(v[i].hi);
       if (t > res) res = t;
    }
 
@@ -104,6 +120,7 @@ static void RowTransformFinish(vec_ZZ& A, quad_float *a, long *in_a)
       }
       else {
          conv(a[i], A(i));
+         CheckFinite(&a[i]);
       }
    }
 }
@@ -116,7 +133,7 @@ static void RowTransformFinish(vec_ZZ& A, quad_float *a, long *in_a)
 
 static void RowTransform(vec_ZZ& A, vec_ZZ& B, const ZZ& MU1, 
                          quad_float *a, quad_float *b, long *in_a, 
-                         quad_float& max_a, quad_float max_b, long& in_float)
+                         double& max_a, double max_b, long& in_float)
 // x = x - y*MU
 {
    static ZZ T, MU;
@@ -128,12 +145,18 @@ static void RowTransform(vec_ZZ& A, vec_ZZ& B, const ZZ& MU1,
    long i;
 
    conv(mu, MU1);
+   CheckFinite(&mu);
 
    if (in_float) {
-      max_a += fabs(mu)*max_b; 
-      if (max_a >= TR_BND) {
+      double mu_abs = fabs(mu);
+      if (mu_abs > 0 && max_b > 0 && (mu_abs >= TR_BND || max_b >= TR_BND)) {
          in_float = 0;
       }
+      else {
+         max_a += mu_abs*max_b;
+         if (max_a >= TR_BND)
+            in_float = 0;
+      }
    }
 
    if (in_float) {
@@ -550,7 +573,7 @@ void GivensComputeGS(quad_float **B1, quad_float **mu, quad_float **aux, long k,
    if (k > n) p[k] = 0;
 
    for (i = 1; i <= k; i++)
-      if (!IsFinite(&p[i])) Error("G_LLL_QP...numbers too big");
+      CheckFinite(&p[i]);
 }
 
 static quad_float red_fudge = to_quad_float(0);
@@ -648,6 +671,7 @@ long ll_G_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
    quad_float mu1;
 
    quad_float t1;
+   double dt1;
    ZZ T1;
    quad_float *tp;
 
@@ -661,8 +685,8 @@ long ll_G_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
    in_vec_mem.SetLength(n+1);
    long *in_vec = in_vec_mem.elts();
 
-   quad_float *max_b;
-   max_b = NTL_NEW_OP quad_float [m+1];
+   double *max_b;
+   max_b = NTL_NEW_OP double [m+1];
    if (!max_b) Error("out of memory in lll_G_LLL_QP");
 
    for (i = 1; i <= m; i++)
@@ -795,7 +819,7 @@ long ll_G_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
             // swap i, i+1
             swap(B(i), B(i+1));
             tp = B1[i]; B1[i] = B1[i+1]; B1[i+1] = tp;
-            t1 = max_b[i]; max_b[i] = max_b[i+1]; max_b[i+1] = t1;
+            dt1 = max_b[i]; max_b[i] = max_b[i+1]; max_b[i+1] = dt1;
             if (U) swap((*U)(i), (*U)(i+1));
          }
 
@@ -823,7 +847,7 @@ long ll_G_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
          // swap rows k, k-1
          swap(B(k), B(k-1));
          tp = B1[k]; B1[k] = B1[k-1]; B1[k-1] = tp;
-         t1 = max_b[k]; max_b[k] = max_b[k-1]; max_b[k-1] = t1;
+         dt1 = max_b[k]; max_b[k] = max_b[k-1]; max_b[k-1] = dt1;
          if (U) swap((*U)(k), (*U)(k-1));
 
          cache.swap();
@@ -900,13 +924,11 @@ long G_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
    }
 
    for (i = 1; i <=m; i++)
-      for (j = 1; j <= n; j++) 
+      for (j = 1; j <= n; j++) {
          conv(B1[i][j], B(i, j));
+         CheckFinite(&B1[i][j]);
+      }
 
-   for (i = 1; i <= m; i++)
-      for (j = 1; j <= n; j++)
-         if (!IsFinite(&B1[i][j]))
-             Error("G_LLL_QP: numbers too big...use G_LLL_XD");
 
    GivensCache_QP cache(m, n);
 
@@ -930,19 +952,19 @@ long G_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
 
    // clean-up
 
-   for (i = 1; i <= m; i++) {
+   for (i = 1; i <= m+dep; i++) {
       delete [] B1[i];
    }
 
    delete [] B1;
 
-   for (i = 1; i <= m; i++) {
+   for (i = 1; i <= m+dep; i++) {
       delete [] mu[i];
    }
 
    delete [] mu;
 
-   for (i = 1; i <= m; i++) {
+   for (i = 1; i <= m+dep; i++) {
       delete [] aux[i];
    }
 
@@ -1064,8 +1086,9 @@ void ComputeG_BKZThresh(quad_float *c, long beta)
 
 
 static 
-void G_BKZStatus(double tt, double enum_time, long NumIterations, 
-               long NumTrivial, long NumNonTrivial, long NumNoOps, long m, 
+void G_BKZStatus(double tt, double enum_time, unsigned long NumIterations, 
+               unsigned long NumTrivial, unsigned long NumNonTrivial, 
+               unsigned long NumNoOps, long m, 
                const mat_ZZ& B)
 {
    cerr << "---- G_BKZ_QP status ----\n";
@@ -1230,8 +1253,7 @@ long G_BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
    for (i = 1; i <=m; i++)
       for (j = 1; j <= n; j++) {
          conv(B1[i][j], B(i, j));
-         if (!IsFinite(&B1[i][j]))
-            Error("G_BKZ_QP: numbers too big...use G_BKZ_XD");
+         CheckFinite(&B1[i][j]);
       }
          
 
@@ -1242,10 +1264,10 @@ long G_BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
    double tt;
 
    double enum_time = 0;
-   long NumIterations = 0;
-   long NumTrivial = 0;
-   long NumNonTrivial = 0;
-   long NumNoOps = 0;
+   unsigned long NumIterations = 0;
+   unsigned long NumTrivial = 0;
+   unsigned long NumNonTrivial = 0;
+   unsigned long NumNoOps = 0;
 
    long verb = verbose;
 
@@ -1302,7 +1324,7 @@ long G_BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
 
          for (i = jj; i <= kk; i++) {
             c[i] = mu[i][i]*mu[i][i];
-            if (!IsFinite(&c[i])) Error("G_BKZ_QP...numbers too big");
+            CheckFinite(&c[i]);
          }
 
          if (prune > 0)
@@ -1464,8 +1486,10 @@ long G_BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                   tp = B1[i-1]; B1[i-1] = B1[i]; B1[i] = tp;
                }
       
-               for (i = 1; i <= n; i++)
+               for (i = 1; i <= n; i++) {
                   conv(B1[jj][i], B(jj, i));
+                  CheckFinite(&B1[jj][i]);
+               }
       
                if (IsZero(B(jj))) Error("G_BKZ_QP: internal error"); 
       
@@ -1562,19 +1586,19 @@ long G_BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
       *UU = *U;
    }
 
-   for (i = 1; i <= m+1; i++) {
+   for (i = 1; i <= m_orig+1; i++) {
       delete [] B1[i];
    }
 
    delete [] B1;
 
-   for (i = 1; i <= m+1; i++) {
+   for (i = 1; i <= m_orig+1; i++) {
       delete [] mu[i];
    }
 
    delete [] mu;
 
-   for (i = 1; i <= m+1; i++) {
+   for (i = 1; i <= m_orig+1; i++) {
       delete [] aux[i];
    }
 
@@ -1628,6 +1652,8 @@ long G_BKZ_QP(mat_ZZ& BB, double delta,
    return G_BKZ_QP(BB, 0, to_quad_float(delta), beta, prune, check);
 }
 
+
+
 static
 long G_BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta, 
          long beta, long prune, LLLCheckFct check)
@@ -1740,8 +1766,7 @@ long G_BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
    for (i = 1; i <=m; i++)
       for (j = 1; j <= n; j++) {
          conv(B1[i][j], B(i, j));
-         if (!IsFinite(&B1[i][j]))
-            Error("G_BKZ_QP: numbers too big...use G_BKZ_XD");
+         CheckFinite(&B1[i][j]);
       }
 
 
@@ -1754,10 +1779,10 @@ long G_BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
    double tt;
 
    double enum_time = 0;
-   long NumIterations = 0;
-   long NumTrivial = 0;
-   long NumNonTrivial = 0;
-   long NumNoOps = 0;
+   unsigned long NumIterations = 0;
+   unsigned long NumTrivial = 0;
+   unsigned long NumNonTrivial = 0;
+   unsigned long NumNoOps = 0;
 
    long verb = verbose;
 
@@ -1814,7 +1839,7 @@ long G_BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
 
          for (i = jj; i <= kk; i++) {
             c[i] = mu[i][i]*mu[i][i];
-            if (!IsFinite(&c[i])) Error("G_BKZ_QP1...numbers too big");
+            CheckFinite(&c[i]);
          }
 
          if (prune > 0)
@@ -1857,6 +1882,8 @@ long G_BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
 
             ctilda[t] = ctilda[t+1] + 
                (yvec[t]+utildavec[t])*(yvec[t]+utildavec[t])*to_double(c[t]);
+
+            ForceToMem(&ctilda[t]); // prevents an infinite loop
    
             if (prune > 0 && t > jj) {
                eta = to_double(G_BKZThresh(t-jj));
@@ -1980,8 +2007,10 @@ long G_BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                   tp = B1[i-1]; B1[i-1] = B1[i]; B1[i] = tp;
                }
       
-               for (i = 1; i <= n; i++)
+               for (i = 1; i <= n; i++) {
                   conv(B1[jj][i], B(jj, i));
+                  CheckFinite(&B1[jj][i]);
+               }
       
                if (IsZero(B(jj))) Error("G_BKZ_QP: internal error"); 
       
@@ -2078,19 +2107,19 @@ long G_BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
       *UU = *U;
    }
 
-   for (i = 1; i <= m+1; i++) {
+   for (i = 1; i <= m_orig+1; i++) {
       delete [] B1[i];
    }
 
    delete [] B1;
 
-   for (i = 1; i <= m+1; i++) {
+   for (i = 1; i <= m_orig+1; i++) {
       delete [] mu[i];
    }
 
    delete [] mu;
 
-   for (i = 1; i <= m+1; i++) {
+   for (i = 1; i <= m_orig+1; i++) {
       delete [] aux[i];
    }
 
