@@ -1065,13 +1065,13 @@ long LLL_FP(mat_ZZ& B, mat_ZZ* U, double delta, long deep,
 
    // clean-up
 
-   for (i = 1; i <= m; i++) {
+   for (i = 1; i <= m+dep; i++) {
       delete [] B1[i];
    }
 
    delete [] B1;
 
-   for (i = 1; i <= m; i++) {
+   for (i = 1; i <= m+dep; i++) {
       delete [] mu[i];
    }
 
@@ -1474,6 +1474,8 @@ long BKZ_FP(mat_ZZ& BB, mat_ZZ* UU, double delta,
 
             ctilda[t] = ctilda[t+1] + 
                (yvec[t]+utildavec[t])*(yvec[t]+utildavec[t])*c[t];
+
+            ForceToMem(&ctilda[t]);  // prevents an infinite loop
    
             if (prune > 0 && t > jj) {
                eta = BKZThresh(t-jj);
@@ -1690,13 +1692,13 @@ long BKZ_FP(mat_ZZ& BB, mat_ZZ* UU, double delta,
       *UU = *U;
    }
 
-   for (i = 1; i <= m+1; i++) {
+   for (i = 1; i <= m_orig+1; i++) {
       delete [] B1[i];
    }
 
    delete [] B1;
 
-   for (i = 1; i <= m+1; i++) {
+   for (i = 1; i <= m_orig+1; i++) {
       delete [] mu[i];
    }
 
