@@ -17,7 +17,7 @@ long zz_pX_gcd_crossover[5] = {400, 400, 800, 1400, 1400};
 long zz_pX_bermass_crossover[5] = {400, 480, 900, 1600, 1600};
 long zz_pX_trace_crossover[5] = {200, 350, 450, 800, 800};
 
-#define QUICK_CRT (NTL_DOUBLE_PRECISION - NTL_SP_NBITS > 10)
+#define QUICK_CRT (NTL_DOUBLE_PRECISION - NTL_SP_NBITS > 12)
 
 
 
@@ -49,9 +49,8 @@ void zz_pX::normalize()
 
    n = rep.length();
    if (n == 0) return;
-   p = rep.elts() + (n-1);
-   while (n > 0 && IsZero(*p)) {
-      p--; 
+   p = rep.elts() + n;
+   while (n > 0 && IsZero(*--p)) {
       n--;
    }
    rep.SetLength(n);
@@ -84,7 +83,7 @@ void SetCoeff(zz_pX& x, long i, zz_p a)
    if (i < 0) 
       Error("SetCoeff: negative index");
 
-   if (i >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (NTL_OVERFLOW(i, 1, 0))
       Error("overflow in SetCoeff");
 
    m = deg(x);
@@ -113,7 +112,7 @@ void SetCoeff(zz_pX& x, long i)
    if (i < 0) 
       Error("coefficient index out of range");
 
-   if (i >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (NTL_OVERFLOW(i, 1, 0))
       Error("overflow in SetCoeff");
 
    m = deg(x);
@@ -356,11 +355,11 @@ void PlainMul(zz_p *xp, const zz_p *ap, long sa, const zz_p *bp, long sb)
 
    for (i = 0; i < sb; i++) {
       long t1 = rep(bp[i]);
-      double bpinv = ((double) t1)*pinv;
+      mulmod_precon_t bpinv = PrepMulModPrecon(t1, p, pinv); // ((double) t1)*pinv;
       zz_p *xp1 = xp+i;
       for (j = 0; j < sa; j++) {
          long t2;
-         t2 = MulMod2(rep(ap[j]), t1, p, bpinv);
+         t2 = MulModPrecon(rep(ap[j]), t1, p, bpinv);
          xp1[j].LoopHole() = AddMod(t2, rep(xp1[j]), p);
       }
    }
@@ -368,7 +367,8 @@ void PlainMul(zz_p *xp, const zz_p *ap, long sa, const zz_p *bp, long sb)
 
 static vec_double a_buf, b_buf;
 
-inline void reduce(zz_p& r, double x, long p, double pinv)
+static inline 
+void reduce(zz_p& r, double x, long p, double pinv)
 {
    long rr = long(x - double(p)*double(long(x*pinv)));
    if (rr < 0) rr += p;
@@ -754,10 +754,10 @@ void PlainSqr(zz_p *xp, const zz_p *ap, long sa)
       const zz_p *ap1 = ap+(j+1);
       zz_p *xp1 = xp+i;
       t1 = rep(ap[j]);
-      double tpinv = ((double) t1)*pinv;
+      mulmod_precon_t tpinv = PrepMulModPrecon(t1, p, pinv); // ((double) t1)*pinv;
 
       for (k = 0; k < cnt; k++) {
-         t2 = MulMod2(rep(ap1[k]), t1, p, tpinv);
+         t2 = MulModPrecon(rep(ap1[k]), t1, p, tpinv);
          t2 = AddMod(t2, rep(xp1[k]), p);
          xp1[k].LoopHole() = t2;
       }
@@ -961,10 +961,10 @@ void PlainDivRem(zz_pX& q, zz_pX& r, const zz_pX& a, const zz_pX& b)
       negate(t, t);
 
       long T = rep(t);
-      double Tpinv = ((double) T)*pinv;
+      mulmod_precon_t Tpinv = PrepMulModPrecon(T, p, pinv); // ((double) T)*pinv;
 
       for (j = db-1; j >= 0; j--) {
-         long S = MulMod2(rep(bp[j]), T, p, Tpinv);
+         long S = MulModPrecon(rep(bp[j]), T, p, Tpinv);
          S = AddMod(S, rep(xp[i+j]), p);
          xp[i+j].LoopHole() = S;
       }
@@ -1039,12 +1039,12 @@ void PlainDiv(zz_pX& q, const zz_pX& a, const zz_pX& b)
       negate(t, t);
 
       long T = rep(t);
-      double Tpinv = ((double) T)*pinv;
+      mulmod_precon_t Tpinv = PrepMulModPrecon(T, p, pinv); // ((double) T)*pinv;
 
       long lastj = max(0, db-i);
 
       for (j = db-1; j >= lastj; j--) {
-         long S = MulMod2(rep(bp[j]), T, p, Tpinv);
+         long S = MulModPrecon(rep(bp[j]), T, p, Tpinv);
          S = AddMod(S, rep(xp[i+j-db]), p);
          xp[i+j-db].LoopHole() = S;
       }
@@ -1102,10 +1102,10 @@ void PlainRem(zz_pX& r, const zz_pX& a, const zz_pX& b)
       negate(t, t);
 
       long T = rep(t);
-      double Tpinv = ((double) T)*pinv;
+      mulmod_precon_t Tpinv = PrepMulModPrecon(T, p, pinv); // ((double) T)*pinv;
 
       for (j = db-1; j >= 0; j--) {
-         long S = MulMod2(rep(bp[j]), T, p, Tpinv);
+         long S = MulModPrecon(rep(bp[j]), T, p, Tpinv);
          S = AddMod(S, rep(xp[i+j]), p);
          xp[i+j].LoopHole() = S;
       }
@@ -1141,7 +1141,7 @@ void mul(zz_pX& x, const zz_pX& a, zz_p b)
    t = rep(b);
    long p = zz_p::modulus();
    double pinv = zz_p::ModulusInverse();
-   double bpinv = t*pinv;
+   mulmod_precon_t bpinv = PrepMulModPrecon(t, p, pinv); // t*pinv;
 
    da = deg(a);
    x.rep.SetLength(da+1);
@@ -1149,7 +1149,7 @@ void mul(zz_pX& x, const zz_pX& a, zz_p b)
    xp = x.rep.elts();
 
    for (i = 0; i <= da; i++) 
-      xp[i].LoopHole() = MulMod2(rep(ap[i]), t, p, bpinv);
+      xp[i].LoopHole() = MulModPrecon(rep(ap[i]), t, p, bpinv);
 
    x.normalize();
 }
@@ -1393,7 +1393,7 @@ void fftRep::SetSize(long NewK)
    n = 1L << NewK;
 
    for (i = 0; i < zz_pInfo->NumPrimes; i++) {
-      if ( !(tbl[i] = (long *) malloc(n * (sizeof (long)))) )
+      if ( !(tbl[i] = (long *) NTL_MALLOC(n, sizeof(long), 0)) )
          Error("out of space in fftRep::SetSize()");
    }
 
@@ -1412,7 +1412,7 @@ fftRep::fftRep(const fftRep& R)
    n = 1L << k;
 
    for (i = 0; i < NumPrimes; i++) {
-      if ( !(tbl[i] = (long *) malloc(n * (sizeof (long)))) )
+      if ( !(tbl[i] = (long *) NTL_MALLOC(n, sizeof(long), 0)) )
          Error("out of space in fftRep");
 
       for (j = 0; j < n; j++)
@@ -1443,7 +1443,7 @@ fftRep& fftRep::operator=(const fftRep& R)
       n = 1L << R.k;
   
       for (i = 0; i < NumPrimes; i++) {
-         if ( !(tbl[i] = (long *) malloc(n * (sizeof (long)))) )
+         if ( !(tbl[i] = (long *) NTL_MALLOC(n, sizeof(long), 0)) )
             Error("out of space in fftRep");
       }
 
@@ -1491,34 +1491,40 @@ void FromModularRep(zz_p& x, long *a)
    long i;
    double y;
 
+
+// I've re-written the following code in v5.3 so that it is 
+// a bit more robust.  
+
 #if QUICK_CRT
+
    y = 0;
    for (i = 0; i < n; i++)
       y = y + ((double) a[i])*zz_pInfo->x[i];
 
-   y = y - long(y*pinv)*p;
-   y = y + 0.5;
+   y = floor(y + 0.5);
+   y = y - floor(y*pinv)*double(p);
    while (y >= p) y -= p;
    while (y < 0) y += p;
    q = long(y);
+
 #else
+
    long Q, r;
-   double qq;
 
    y = 0;
-   qq = 0;
+   q = 0;
 
    for (i = 0; i < n; i++) {
       r = MulDivRem(Q, a[i], zz_pInfo->u[i], FFTPrime[i], zz_pInfo->x[i]);
-      qq = qq + Q;
+      q = q + Q;
+#if (NTL_BITS_PER_LONG - NTL_SP_NBITS <= 4)
+      // on typical platforms, this reduction will not be necessary.
+      q = q % p;
+#endif
       y = y + r*FFTPrimeInv[i];
    }
 
-   y = qq + long(y + 0.5);
-   y = y - long(y*pinv)*p;
-   while (y >= p) y -= p;
-   while (y < 0) y += p;
-   q = long(y); 
+   q = (q + long(y + 0.5)) % p;
 
 #endif
 
@@ -1535,6 +1541,7 @@ void FromModularRep(zz_p& x, long *a)
 }
 
 
+
 void TofftRep(fftRep& y, const zz_pX& x, long k, long lo, long hi)
 // computes an n = 2^k point convolution.
 // if deg(x) >= 2^k, then x is first reduced modulo X^n-1.
@@ -1725,7 +1732,7 @@ void FromfftRep(zz_pX& x, fftRep& y, long lo, long hi)
    long NumPrimes = zz_pInfo->NumPrimes;
 
    long t[4];
-   vec_long& s = FFTBuf;;
+   vec_long& s = FFTBuf;
 
    k = y.k;
    n = (1L << k);
@@ -2759,7 +2766,7 @@ void InvTrunc(zz_pX& x, const zz_pX& a, long m)
       return;
    }
 
-   if (m >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (NTL_OVERFLOW(m, 1, 0))
       Error("overflow in InvTrunc");
 
    if (&x == &a) {
@@ -3292,8 +3299,8 @@ void power(zz_pX& x, const zz_pX& a, long e)
 
 void reverse(zz_pX& x, const zz_pX& a, long hi)
 {
-   if (hi < -1) Error("reverse: bad args");
-   if (hi >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (hi < 0) { clear(x); return; }
+   if (NTL_OVERFLOW(hi, 1, 0))
       Error("overflow in reverse");
 
    if (&x == &a) {
